#include <stdio.h> /* FILE operations */
#include <stdlib.h>
#include <math.h> /* For log(), for giving dB results */
#include <unistd.h>
#include <stdlib.h> /* For exit() */
#include <getopt.h> /* ???????? I think getopt() only requires unistd! */
#include <string.h> /* strncpy() */
#include <gsl/gsl_matrix.h> /* We need matrix operations for matrices */
#include "libjpegrest.h"

/* imgerror. This program gives quality measures of images.
*/
enum global_options
{
	FILENAME_LENGTH=200
};


/* Global commandline options */
enum e_filetype
{
	TYPE_JPEG,
	TYPE_PAM
};

enum e_action
{
	ACTION_UNKNOWN,
	ACTION_GBIM,
	ACTION_STATISTIC,
	ACTION_PSNR,
	ACTION_SHOWALL
};

struct {
	char filename[2][FILENAME_LENGTH];
	enum e_filetype filetype[2];
	int nfiles;
	enum e_action action;
	int plane;
	int rgb2ycc; /* boolean */
} commandline;


void show_measures(const struct bitmap bmap[2])
{
	int plane, ifile;

	switch(commandline.action)
	{
	case ACTION_GBIM:
		/* Only one file supported, in single GBIM: bmap[0]. */
		for (plane = 0; plane < bmap[0].nplanes; plane++)
		{
			printf("%g", error_mgbim(bmap[0].m[plane]));
			/* Space split the values */
			if (plane != bmap[0].nplanes - 1)
				printf(" ");
		}
		printf("\n");
		break;
	case ACTION_STATISTIC:
		/* Only one file supported, in single STATISTIC: bmap[0]. */
		for (plane = 0; plane < bmap[0].nplanes; plane++)
		{
			double stddev;
			double mean;

			error_statistic(bmap[0].m[plane], &mean,
					&stddev);
			printf("%g %g\n", mean, stddev);
		}
		break;
	case ACTION_SHOWALL:
		for (ifile = 0; ifile < commandline.nfiles; ifile++)
		{
			printf("Filename: %s\n", commandline.filename[ifile]);
			printf(" Number of planes: %i\n", bmap[ifile].nplanes);
			for (plane = 0; plane < bmap[ifile].nplanes; plane++)
			{
				printf(" MGBIM[%i]: %g\n", plane,
						error_mgbim(
						bmap[ifile].m[plane]));
				fflush(stdout);
			}
			/* Only one file supported, in single STATISTIC: bmap[0]. */
			for (plane = 0; plane < bmap[ifile].nplanes; plane++)
			{
				double stddev;
				double mean;
				error_statistic(bmap[ifile].m[plane], &mean,
						&stddev);
				printf(" MEAN[%i]: %g\n", plane, mean);
				printf(" STDDEV[%i]: %g\n", plane, stddev);
			}
		}
		if (commandline.nfiles != 2)
			break;
		else if (!are_bitmaps_comparable(bmap[0], bmap[1]))
		{
			fprintf(stderr, "The two given bitmaps aren't "
					"comparable.\n");
		}
		else
		{
			/* We use the first bitmap as plane reference */
			for (plane = 0; plane < bmap[0].nplanes; plane++)
			{
				printf( "PSNR[%i]: %g dB\n", plane,
					20 * log(error_psnr(bmap[0].m[plane],
						bmap[1].m[plane])) );
				fflush(stdout);
			}
		}
		break;
	case ACTION_PSNR:
		if (commandline.nfiles != 2)
		{
			fprintf(stderr, "PSNR cannot be calculated "
					"for a single file.\n");
		} else if (!are_bitmaps_comparable(bmap[0], bmap[1]))
		{
			fprintf(stderr, "The two given bitmaps aren't "
					"comparable.\n");
		}
		else
		{
			for (plane = 0; plane < bmap[0].nplanes; plane++)
			{
				printf( "%g",
					error_psnr(bmap[0].m[plane],
						bmap[1].m[plane]) );
				/* Space split the values */
				if (plane != bmap[0].nplanes - 1)
					printf(" ");
			}
			printf("\n");
		}
		break;
	default:
		fprintf(stderr, "I don't know what to show.\n");
	}
}

void parse_commandline(int *argc, char *argv[])
{
	char option;
	int error;
	int showhelp;


	commandline.plane = 0;
	commandline.rgb2ycc = 0; /* False */
	commandline.action = ACTION_SHOWALL;
	error = 0;
	showhelp = 0;

	while((option = getopt(*argc, argv, "j:p:gsty")) != -1)
	{
		switch(option)
		{
		case 'j':
			if (*optarg == ':')
			{
				fprintf(stderr, "Error: -j needs a filename\n");
				showhelp = 1;
				error = 1;
			} else
			{
				strncpy(commandline.filename[
						commandline.nfiles],
						optarg, FILENAME_LENGTH);
				commandline.filetype[commandline.nfiles] =
					TYPE_JPEG;
				commandline.nfiles++;
			}
			break;
		case 'p':
			if (*optarg == ':')
			{
				fprintf(stderr, "Error: -p needs a filename\n");
				showhelp = 1;
				error = 1;
			} else
			{
				strncpy(commandline.filename[
						commandline.nfiles],
						optarg, FILENAME_LENGTH);
				commandline.filetype[commandline.nfiles] =
					TYPE_PAM;
				commandline.nfiles++;
			}
			break;
		case 'g':
			commandline.action = ACTION_GBIM;
			break;
		case 't':
			commandline.action = ACTION_STATISTIC;
			break;
		case 's':
			commandline.action = ACTION_PSNR;
			break;
		case 'y':
			commandline.rgb2ycc = 1; /* True */
			break;
		case '?':
		default:
			fprintf(stderr, "Unknown command line option\n");
			showhelp = 1;
			error = 1;
		}
	}

	if (commandline.nfiles == 0)
	{
		showhelp = 1;
		error = 1;
	} else if (commandline.action == ACTION_GBIM && commandline.nfiles == 2)
	{
		fprintf(stderr,
			"If you want only GBIM, give only a single file.\n");
		showhelp = 0;
		error = 1;
	} else if (commandline.action == ACTION_PSNR && commandline.nfiles < 2)
	{
		fprintf(stderr, "If you want only PSNR, give two files.\n");
		showhelp = 0;
		error = 1;
	}

	if (showhelp)
	{
		fprintf(stderr,
"usage: %s [-gsty] [-j <image.jpeg>] [-p <image.pnm>]\n"
"  -g   show MGBIM only.\n"
"  -s   show PSNR only as a float in linear scale.\n"
"  -t   show the mean and the stddev values for each plane.\n"
"  -y   measure the YCbCr planes instead of RGB.\n"
"  -j   open a JPEG image.\n"
"  -p   open a PAM image.\n"
"  without -s or -j : show all the error measures in human readable format.\n"
"  In PSNR, the first image given is the reference, which the second is\n"
"  compared to.\n"
			, argv[0]);
	}

	if (error)
	{
		exit(-1);
	}
}

void get_files(struct bitmap bmap[2])
{
	int ifile;
	int result;

	for(ifile = 0; ifile < commandline.nfiles; ifile++)
	{
		switch(commandline.filetype[ifile])
		{
		case TYPE_PAM:
			result = bitmap_filename_pam_get(
					commandline.filename[ifile],
					&bmap[ifile]);
			if(result == -1)
				exit(-1);
			break;
		case TYPE_JPEG:
			result = bitmap_filename_jpeg_get(
					commandline.filename[ifile],
					&bmap[ifile], READ_RGB);
			if(result == -1)
				exit(-1);
			break;
		}

		if (commandline.rgb2ycc && bmap[ifile].nplanes == 3)
			rgb2ycc(bmap[ifile].m[0],
					bmap[ifile].m[1],
					bmap[ifile].m[2]);
	}
}

int main(int argc, char *argv[])
{

	int i;

	struct bitmap bmap[2];

	parse_commandline(&argc, argv);
	bitmap_pam_init(&argc, argv);
	bitmap_jpeg_init(&argc, argv);

	/* Get the data from the file */
	get_files(bmap);

	show_measures(bmap);
	
	for (i=0; i<commandline.nfiles; i++)
		bitmap_free(bmap[i]);

	return 0;
}
